
import sys
import abc
import io
import time
import traceback
import math
import copy
import pickle
import numpy as np
import torch
import scipy.optimize

class SVEM(abc.ABC):

    @abc.abstractmethod
    def maximize(self, model, optim_params, method="ECM", getIterationModelParamsFn=None,
                 printIterationModelParams=True,
                 logLock=None, logStreamFN=None,
                 lowerBoundLock=None, lowerBoundStreamFN=None,
                 latentsTimes=None, latentsLock=None, latentsStreamFN=None,
                 verbose=True, out=sys.stdout,
                 savePartial=False,
                 savePartialFilenamePattern="results/00000000_{:s}_estimatedModel.pickle",
                ):
        """Maximizes the sparse variational lower bound, Eq. 4 in Duncker and
        Sahani, 2018.

        .. note::
            Only parameters **model**, **optim_params** and **method** should
            be set by users. The remaining parameters are used to interface
            :meth:`svGPFA.stats.svEM.SVEM.maximize` with the dashboard.

        :param model: svGPFA model used to calculate the lower bound
        :type  model: :class:`svGPFA.stats.svLowerBound.SVLowerBound`
        :param optim_params: optimization parameters. The format of this dictionary is identical to the one described in :any:`optim_params`. It can be obtained at the key ``optim_params`` of the dictionary returned by :func:`svGPFA.utils.initUtils.getParamsAndKernelsTypes`.
        :type  optim_params: dictionary
        :param method: either ECM for Expectation Conditional Maximization or mECM for multicycle expectation conditional maximization. Refer to :cite:t:`mcLachlanAndKrishnan08` for details on these algorithms.
        """
        pass


    @abc.abstractmethod
    def _eStep(self, model, optim_params):
        pass


    @abc.abstractmethod
    def _mStepEmbedding(self, model, optim_params):
        pass


    @abc.abstractmethod
    def _mStepKernels(self, model, optim_params):
        pass


    @abc.abstractmethod
    def _mStepIndPointsLocs(self, model, optim_params):
        pass


    def _writeToLockedLog(self, message, logLock, logStream, logStreamFN):
        logStream.write(message)
        if logLock is not None and not logLock.is_locked():
            logLock.lock()
            with open(logStreamFN, 'a') as f:
                f.write(logStream.getvalue())
            logLock.unlock()
            logStream.truncate(0)
            logStream.seek(0)


class SVEM_PyTorch(SVEM):

    def maximize(self, model, optim_params, method="ECM", getIterationModelParamsFn=None,
                 printIterationModelParams=True,
                 logLock=None, logStreamFN=None,
                 lowerBoundLock=None, lowerBoundStreamFN=None,
                 latentsTimes=None, latentsLock=None, latentsStreamFN=None,
                 verbose=True, out=sys.stdout,
                 savePartial=False,
                 savePartialFilenamePattern="results/00000000_{:s}_estimatedModel.pickle",
                ):
        if latentsStreamFN is not None and latentsTimes is None:
            raise RuntimeError("Please specify latentsTime if you want to save latents")

        iter = 0
        if savePartial:
            savePartialFilename = savePartialFilenamePattern.format("initial")
            resultsToSave = {"model": model}
            with open(savePartialFilename, "wb") as f: pickle.dump(resultsToSave, f)
        lowerBound0 = model.eval()
        lowerBoundHist = [lowerBound0.item()]
        elapsedTimeHist = [0.0]
        startTime = time.time()

        if lowerBoundLock is not None and lowerBoundStreamFN is not None and not lowerBoundLock.is_locked():
            lowerBoundLock.lock()
            with open(lowerBoundStreamFN, 'wb') as f:
                np.save(f, np.array(lowerBoundHist))
            lowerBoundLock.unlock()

        if latentsLock is not None and latentsStreamFN is not None and not latentsLock.is_locked():
            latentsLock.lock()
            muK, varK = model.predictLatents(newTimes=latentsTimes)

            with open(latentsStreamFN, 'wb') as f:
                np.savez(f, iteration=iter, times=latentsTimes.detach().numpy(), muK=muK.detach().numpy(), varK=varK.detach().numpy())
            lowerBoundLock.unlock()
        logStream = io.StringIO()
        if method.lower() == "ecm":
            steps = ["estep", "mstep_embedding", "mstep_kernels",
                     "mstep_indpointslocs"]
        elif method.lower() == "mecm":
            # see McLachlan, G. J., & Krishnan, T. (2007). The EM algorithm and
            # extensions (Vol. 382). John Wiley & Sons) Chapter 5
            steps = ["estep", "mstep_embedding", "estep", "mstep_kernels",
                     "estep", "mstep_indpointslocs"]
        else:
            raise ValueError("Invalid method={:s}. Supported values are ECM and mECM".format(method))
        functions_for_steps = {"estep": self._eStep,
                               "mstep_embedding": self._mStepEmbedding,
                               "mstep_kernels": self._mStepKernels,
                               "mstep_indpointslocs": self._mStepIndPointsLocs}
        if getIterationModelParamsFn is not None:
            iterationModelParams = getIterationModelParamsFn(model=model)
            if printIterationModelParams:
                print(iterationModelParams)
            iterationsModelParams = [None for i in
                                     range(optim_params["em_max_iter"]+1)]
            iterationsModelParams[iter] = iterationModelParams
        else:
            iterationsModelParams = None
        maxRes = {"lowerBound": -math.inf}
        iter += 1
        while iter <= optim_params["em_max_iter"]:
            for step in steps:
                if optim_params["{:s}_estimate".format(step)]:
                    message = "Iteration {:02d}, {:s} start: {:f}\n".format(
                        iter, step, maxRes["lowerBound"])
                    if verbose:
                        out.write(message)
                    self._writeToLockedLog(
                        message=message,
                        logLock=logLock,
                        logStream=logStream,
                        logStreamFN=logStreamFN
                    )
                    try:
                        maxRes = functions_for_steps[step](
                            model=model,
                            optim_params=optim_params["{:s}_optim_params".format(
                                step)])
                    except Exception as e:
                        stack_trace = traceback.format_exc()
                        print(e)
                        print(stack_trace)
                        terminationInfo = ErrorTerminationInfo(
                            message=f"Error occured while processing {step} in iteration {iter}",
                            error=e, stack_trace=stack_trace)
                        return lowerBoundHist, elapsedTimeHist, \
                               terminationInfo, iterationsModelParams
                    message = "Iteration {:02d}, {:s} end: {:f}, niter: {:d}, nfeval: {:d}\n".format(
                        iter, step, maxRes["lowerBound"], maxRes["niter"],
                        maxRes["nfeval"])
                    if verbose:
                        out.write(message)
                    self._writeToLockedLog(
                        message=message,
                        logLock=logLock,
                        logStream=logStream,
                        logStreamFN=logStreamFN
                    )
                    if savePartial:
                        savePartialFilename = \
                            savePartialFilenamePattern.format(
                                "{:s}{:03d}".format(step, iter))
                        resultsToSave = {"model": model}
                        with open(savePartialFilename, "wb") as f:
                            pickle.dump(resultsToSave, f)
                    if getIterationModelParamsFn is not None:
                        iterationModelParams = getIterationModelParamsFn(model=model)
                        if printIterationModelParams:
                            print(iterationModelParams)
                        iterationsModelParams[iter] = iterationModelParams
            elapsedTimeHist.append(time.time()-startTime)
            lowerBoundHist.append(maxRes["lowerBound"].item())

            if lowerBoundLock is not None and \
               lowerBoundStreamFN is not None and \
               not lowerBoundLock.is_locked():
                lowerBoundLock.lock()
                with open(lowerBoundStreamFN, 'wb') as f:
                    np.save(f, np.array(lowerBoundHist))
                lowerBoundLock.unlock()

            if latentsLock is not None and \
               latentsStreamFN is not None and \
               not latentsLock.is_locked():
                latentsLock.lock()
                muK, varK = model.predictLatents(newTimes=latentsTimes)

                with open(latentsStreamFN, 'wb') as f:
                    np.savez(f, iteration=iter,
                             times=latentsTimes.detach().numpy(),
                             muK=muK.detach().numpy(),
                             varK=varK.detach().numpy())
                lowerBoundLock.unlock()

            iter += 1
        terminationInfo = TerminationInfo(
            "Maximum number of iterations ({:d}) reached".format(
                optim_params["em_max_iter"]))
        return lowerBoundHist, elapsedTimeHist, terminationInfo, \
            iterationsModelParams

    def _eStep(self, model, optim_params):
        x = model.getSVPosteriorOnIndPointsParams()
        evalFunc = model.eval
        optimizer = torch.optim.LBFGS(x, **optim_params)
        answer = self._setupAndMaximizeStep(x=x, evalFunc=evalFunc, optimizer=optimizer)
        return answer

    def _mStepEmbedding(self, model, optim_params):
        x = model.getSVEmbeddingParams()
        svPosteriorOnLatentsStats = model.computeSVPosteriorOnLatentsStats()
        evalFunc = lambda: model.evalELLSumAcrossTrialsAndNeurons(svPosteriorOnLatentsStats=svPosteriorOnLatentsStats)
        # evalFunc = model.eval
        optimizer = torch.optim.LBFGS(x, **optim_params)
        answer = self._setupAndMaximizeStep(x=x, evalFunc=evalFunc, optimizer=optimizer)
        return answer

    def _mStepKernels(self, model, optim_params):
        x = model.getKernelsParams()
        prev_x = [copy.deepcopy(anx) for anx in x]
        def evalFunc():
            model.buildKernelsMatrices()
            answer = model.eval()
            return answer
        optimizer = torch.optim.LBFGS(x, **optim_params)
        try:
            answer = self._setupAndMaximizeStep(x=x, evalFunc=evalFunc, optimizer=optimizer)
            # begin debug
            # print("*** Updated kernels params ***")
            # print(x)
            # raise ValueError("Test error in _mStepKernels")
            # end debug
        except Exception:
            x = [anx.detach() for anx in x]
            for i in range(len(x)):
                x[i][:] = prev_x[i][:]
            raise
        return answer

    def _mStepIndPointsLocs(self, model, optim_params):
        x = model.getIndPointsLocs()
        def evalFunc():
            model.buildKernelsMatrices()
            answer = model.eval()
            return answer
        optimizer = torch.optim.LBFGS(x, **optim_params)
        answer = self._setupAndMaximizeStep(x=x, evalFunc=evalFunc, optimizer=optimizer)
        return answer

    def _setupAndMaximizeStep(self, x, evalFunc, optimizer):
        for i in range(len(x)):
            x[i].requires_grad = True
        maxRes = self._maximizeStep(evalFunc=evalFunc, optimizer=optimizer)
        for i in range(len(x)):
            x[i].requires_grad = False
        return maxRes

    def _maximizeStep(self, evalFunc, optimizer):
        def closure():
            optimizer.zero_grad()
            curEval = -evalFunc()
            # begin debug
            # print("before -- eval {:f}, params ".format(curEval), self._model.getKernelsParams())
            # end debug
            curEval.backward(retain_graph=True)
            # begin debug
            # with torch.no_grad():
            #     auxEval = -evalFunc()
            # print("after -- eval {:f}, params ".format(auxEval), self._model.getKernelsParams())
            # print("after -- params ", self._model.getKernelsParams())
            # end debug
            return curEval
        optimizer.step(closure)
        lowerBound = evalFunc()
        stateOneEpoch = optimizer.state[optimizer._params[0]]
        nfeval = stateOneEpoch["func_evals"]
        niter = stateOneEpoch["n_iter"]
        return {"lowerBound": lowerBound, "nfeval": nfeval, "niter": niter}

class SVEM_SciPy(SVEM):

    def maximize(self, model, optim_params, method="EM", getIterationModelParamsFn=None,
                 printIterationModelParams=True,
                 logLock=None, logStreamFN=None,
                 lowerBoundLock=None, lowerBoundStreamFN=None,
                 latentsTimes=None, latentsLock=None, latentsStreamFN=None,
                 verbose=True, out=sys.stdout,
                 savePartial=False,
                 savePartialFilenamePattern="00000000_{:s}_estimatedModel.pickle",
                ):

        if latentsStreamFN is not None and latentsTimes is None:
            raise RuntimeError("Please specify latentsTime if you want to save latents")

        iter = 0
        if savePartial:
            savePartialFilename = savePartialFilenamePattern.format("initial")
            resultsToSave = {"model": model}
            with open(savePartialFilename, "wb") as f: pickle.dump(resultsToSave, f)
        lowerBound0 = model.eval()
        lowerBoundHist = [lowerBound0.item()]
        elapsedTimeHist = [0.0]
        startTime = time.time()

        if lowerBoundLock is not None and lowerBoundStreamFN is not None and not lowerBoundLock.is_locked():
            lowerBoundLock.lock()
            with open(lowerBoundStreamFN, 'wb') as f:
                np.save(f, np.array(lowerBoundHist))
            lowerBoundLock.unlock()

        if latentsLock is not None and latentsStreamFN is not None and not latentsLock.is_locked():
            latentsLock.lock()
            muK, varK = model.predictLatents(newTimes=latentsTimes)

            with open(latentsStreamFN, 'wb') as f:
                np.savez(f, iteration=iter, times=latentsTimes.detach().numpy(), muK=muK.detach().numpy(), varK=varK.detach().numpy())
            lowerBoundLock.unlock()
        iter += 1
        logStream = io.StringIO()
        if method=="EM":
            steps = ["estep", "mstep_embedding", "mstep_kernels", "mstep_indpointslocs"]
            functions_for_steps = {"estep": self._eStep, "mstep_embedding": self._mStepEmbedding, "mstep_kernels": self._mStepKernels, "mstep_indpointslocs": self._mStepIndPointsLocs}
        elif method=="mECM":
            steps = ["estep", "mstep_embedding", "estep", "mstep_kernels", "estep", "mstep_indpointslocs"]
            functions_for_steps = {"estep": self._eStep, "mstep_embedding": self._mStepEmbedding, "estep": self._eStep, "mstep_kernels": self._mStepKernels, "estep": self._eStep, "mstep_indpointslocs": self._mStepIndPointsLocs}
        else:
            raise ValueError("Invalid method={:s}. Supported values are EM and mECM".format(method))
        if getIterationModelParamsFn is not None:
            initialModelsParams = getIterationModelParamsFn(model=model)
            iterationsModelParams = torch.empty((optim_params["em_max_iter"]+1, len(initialModelsParams)), dtype=torch.double)
            iterationsModelParams[0,:] = initialModelsParams
        else:
            iterationsModelParams = None
        maxRes = {"lowerBound": -math.inf}
        while iter<optim_params["em_max_iter"]:
            for step in steps:
                if optim_params["{:s}_estimate".format(step)]:
                    message = "Iteration {:02d}, {:s} start: {:f}\n".format(iter, step, maxRes["lowerBound"])
                    if verbose:
                        out.write(message)
                    self._writeToLockedLog(
                        message=message,
                        logLock=logLock,
                        logStream=logStream,
                        logStreamFN=logStreamFN
                    )
                    maxRes = functions_for_steps[step](model=model, optim_params=optim_params["{:s}_optim_params".format(step)])
                    message = "Iteration {:02d}, {:s} end: {:f}, niter: {:d}, nfeval: {:d}, success: {:d}\n".format(iter, step, maxRes["lowerBound"], maxRes["niter"], maxRes["nfeval"], maxRes["success"])
                    if verbose:
                        out.write(message)
                    self._writeToLockedLog(
                        message=message,
                        logLock=logLock,
                        logStream=logStream,
                        logStreamFN=logStreamFN
                    )
                    if savePartial:
                        savePartialFilename = savePartialFilenamePattern.format("{:s}{:03d}".format(step, iter))
                        resultsToSave = {"model": model}
                        with open(savePartialFilename, "wb") as f: pickle.dump(resultsToSave, f)
                    if getIterationModelParamsFn is not None:
                        iterationsModelParams[iter+1,:] = getIterationModelParamsFn(model=model)
            elapsedTimeHist.append(time.time()-startTime)
            lowerBoundHist.append(maxRes)

            if lowerBoundLock is not None and lowerBoundStreamFN is not None and not lowerBoundLock.is_locked():
                lowerBoundLock.lock()
                with open(lowerBoundStreamFN, 'wb') as f:
                    np.save(f, np.array(lowerBoundHist))
                lowerBoundLock.unlock()

            if latentsLock is not None and latentsStreamFN is not None and not latentsLock.is_locked():
                latentsLock.lock()
                muK, varK = model.predictLatents(newTimes=latentsTimes)

                with open(latentsStreamFN, 'wb') as f:
                    np.savez(f, iteration=iter, times=latentsTimes.detach().numpy(), muK=muK.detach().numpy(), varK=varK.detach().numpy())
                lowerBoundLock.unlock()

            iter += 1
        terminationInfo = TerminationInfo("Maximum number of iterations ({:d}) reached".format(optim_params["em_max_iter"]))
        return lowerBoundHist, elapsedTimeHist, terminationInfo, iterationsModelParams

    def _eStep(self, model, optim_params, method="L-BFGS-B"):
        def eval_func(z):
            model.set_svPosteriorOnIndPoints_params_from_flattened(flattened_params=z.tolist())
            model.set_svPosteriorOnIndPoints_params_requires_grad(requires_grad=True)
            value = -model.eval()
            value.backward(retain_graph=True)
            grad_list = model.get_flattened_svPosteriorOnIndPoints_params_grad()
            value = value.item()
            grad = np.array(grad_list)
            return (value, grad)

        z0 = np.array(model.get_flattened_svPosteriorOnIndPoints_params())
        optim_res = scipy.optimize.minimize(fun=eval_func, x0=z0,
                                            method=method, jac=True,
                                            options=optim_params)
        model.set_svPosteriorOnIndPoints_params_requires_grad(requires_grad=False)
        model.set_svPosteriorOnIndPoints_params_from_flattened(flattened_params=optim_res.x.tolist())
        lowerBound = -optim_res.fun
        nfeval = optim_res.nfev
        niter = optim_res.nit
        success = optim_res.success
        answer = {"lowerBound": lowerBound, "nfeval": nfeval, "niter": niter, "success": success}

    def _eStep(self, model, optim_params, method="L-BFGS-B"):
        def eval_func(z):
            model.set_svPosteriorOnIndPoints_params_from_flattened(flattened_params=z.tolist())
            model.set_svPosteriorOnIndPoints_params_requires_grad(requires_grad=True)
            value = -model.eval()
            value.backward(retain_graph=True)
            grad_list = model.get_flattened_svPosteriorOnIndPoints_params_grad()
            value = value.item()
            grad = np.array(grad_list)
            return (value, grad)

        z0 = np.array(model.get_flattened_svPosteriorOnIndPoints_params())
        optim_res = scipy.optimize.minimize(fun=eval_func, x0=z0,
                                            method=method, jac=True,
                                            options=optim_params)
        model.set_svPosteriorOnIndPoints_params_requires_grad(requires_grad=False)
        model.set_svPosteriorOnIndPoints_params_from_flattened(flattened_params=optim_res.x.tolist())
        lowerBound = -optim_res.fun
        nfeval = optim_res.nfev
        niter = optim_res.nit
        success = optim_res.success
        answer = {"lowerBound": lowerBound, "nfeval": nfeval, "niter": niter, "success": success}
        return answer

    def _mStepEmbedding(self, model, optim_params, method="L-BFGS-B"):
        def eval_func(z):
            model.set_svEmbedding_params_from_flattened(flattened_params=z.tolist())
            model.set_svEmbedding_params_requires_grad(requires_grad=True)
#             svPosteriorOnLatentsStats = model.computeSVPosteriorOnLatentsStats()
#             value = -model.evalELLSumAcrossTrialsAndNeurons(svPosteriorOnLatentsStats=svPosteriorOnLatentsStats)
            value = -model.eval()
            value.backward(retain_graph=True)
            grad_list = model.get_flattened_svEmbedding_params_grad()
            value = value.item()
            grad = np.array(grad_list)
            return (value, grad)

        z0 = np.array(model.get_flattened_svEmbedding_params())
        optim_res = scipy.optimize.minimize(fun=eval_func, x0=z0,
                                            method=method, jac=True,
                                            options=optim_params)
        model.set_svEmbedding_params_requires_grad(requires_grad=False)
        model.set_svEmbedding_params_from_flattened(flattened_params=optim_res.x.tolist())
        lowerBound = -optim_res.fun
        nfeval = optim_res.nfev
        niter = optim_res.nit
        success = optim_res.success
        answer = {"lowerBound": lowerBound, "nfeval": nfeval, "niter": niter, "success": success}
        return answer

    def _mStepKernels(self, model, optim_params, method="L-BFGS-B"):
        def eval_func(z):
            model.set_kernels_params_from_flattened(flattened_params=z.tolist())
            model.set_kernels_params_requires_grad(requires_grad=True)
            model.buildKernelsMatrices()
            value = -model.eval()
            value.backward(retain_graph=True)
            grad_list = model.get_flattened_kernels_params_grad()
            value = value.item()
            grad = np.array(grad_list)
            return (value, grad)

        z0 = np.array(model.get_flattened_kernels_params())
        optim_res = scipy.optimize.minimize(fun=eval_func, x0=z0,
                                            method=method, jac=True,
                                            # bounds=scipy.optimize.Bounds(lb=0.1, ub=30.0),
                                            options=optim_params)
        model.set_kernels_params_requires_grad(requires_grad=False)
        model.set_kernels_params_from_flattened(flattened_params=optim_res.x.tolist())
        lowerBound = -optim_res.fun
        # print("*** kernels parameters: {}".format(model.getKernelsParams()))
        print("*** kernels parameters: {}".format(optim_res.x))
        nfeval = optim_res.nfev
        niter = optim_res.nit
        success = optim_res.success
        answer = {"lowerBound": lowerBound, "nfeval": nfeval, "niter": niter, "success": success}
        return answer

    def _mStepIndPointsLocs(self, model, optim_params, method="L-BFGS-B"):
        def eval_func(z):
            model.set_indPointsLocs_from_flattened(flattened_params=z.tolist())
            model.set_indPointsLocs_requires_grad(requires_grad=True)
            model.buildKernelsMatrices()
            value = -model.eval()
            value.backward(retain_graph=True)
            grad_list = model.get_flattened_indPointsLocs_grad()
            value = value.item()
            grad = np.array(grad_list)
            return (value, grad)

        z0 = np.array(model.get_flattened_indPointsLocs())
        optim_res = scipy.optimize.minimize(fun=eval_func, x0=z0,
                                            method=method, jac=True,
                                            options=optim_params)
        model.set_indPointsLocs_requires_grad(requires_grad=False)
        model.set_indPointsLocs_from_flattened(flattened_params=optim_res.x.tolist())
        lowerBound = -optim_res.fun
        nfeval = optim_res.nfev
        niter = optim_res.nit
        success = optim_res.success
        answer = {"lowerBound": lowerBound, "nfeval": nfeval, "niter": niter, "success": success}
        return answer

class TerminationInfo:
    def __init__(self, message):
        self._message = message

    @property
    def message(self):
        return self._message

class ErrorTerminationInfo(TerminationInfo):
    def __init__(self, message, error, stack_trace):
        super().__init__(message=message)
        self._error = error
        self._stack_trace = stack_trace

    @property
    def error(self):
        return self._error

    @property
    def stack_trace(self):
        return self._stack_trace




import scipy.io
import math
import numpy as np
import pandas as pd
import scipy
import sklearn.metrics
import numpy as np
import torch
import warnings

# from . import my_globals
import svGPFA.stats.kernels
import gcnu_common.numerical_methods.utils
import gcnu_common.stats.gaussianProcesses.eval


def separateNeuronsSpikeTimesByTrials(neurons_spike_times, epochs_times,
                                      trials_start_times_rel,
                                      trials_end_times_rel):
    n_trials = len(epochs_times)
    n_neurons = len(neurons_spike_times)
    trials_spikes_times = [[] for r in range(n_trials)]
    for r in range(n_trials):
        trial_epoch_time = epochs_times[r]
        trial_start_time_rel = trials_start_times_rel[r]
        trial_end_time_rel = trials_end_times_rel[r]
        trial_spikes_times = [[] for n in range(n_neurons)]
        for n in range(n_neurons):
            neuron_spikes_times_rel = neurons_spike_times[n]-trial_epoch_time
            trial_neuron_spikes_times = neuron_spikes_times_rel[
                np.logical_and(trial_start_time_rel <= neuron_spikes_times_rel,
                               neuron_spikes_times_rel < trial_end_time_rel)]
            trial_spikes_times[n] = trial_neuron_spikes_times
        trials_spikes_times[r] = trial_spikes_times
    return trials_spikes_times


def buildKernels(kernels_types, kernels_params):
    n_latents = len(kernels_types)
    kernels = [None for k in range(n_latents)]

    for k, kernel_type in enumerate(kernels_types):
        if kernels_types[k] == "exponentialQuadratic":
            kernels[k] = svGPFA.stats.kernels.ExponentialQuadraticKernel()
        elif kernels_types[k] == "periodic":
            kernels[k] = svGPFA.stats.kernels.PeriodicKernel()
        else:
            raise ValueError(f"Invalid kernels type: {kernels_types[k]}")
        kernels[k].setParams(kernels_params[k])
    return kernels


def orthonormalizeLatentsMeans(latents_means, C):
    U, S, Vh = np.linalg.svd(C)
    orthoMatrix = Vh.T*S
    n_trials = len(latents_means)
    oLatentsMeans = [[] for r in range(n_trials)]
    for r in range(n_trials):
        oLatentsMeans[r] = np.matmul(latents_means[r], orthoMatrix)
    return oLatentsMeans


def getPropSamplesCovered(sample, mean, std, percent=.95):
    if percent==.95:
        factor = 1.96
    else:
        raise ValueError("percent=0.95 is the only option implemented at the moment")
    covered = torch.logical_and(mean-factor*std<=sample, sample<mean+factor*std)
    coverage = torch.count_nonzero(covered)/float(len(covered))
    return coverage

def getCIFs(C, d, latents):
    n_trials = latents.shape[0]
    nLatents = latents.shape[2]
    nSamples = latents.shape[1]
    nNeurons = C.shape[0]
    embeddings = torch.empty((n_trials, nSamples, nNeurons))
    for r in range(n_trials):
        embeddings[r,:,:] = torch.matmul(latents[r,:,:], torch.transpose(C, 0, 1))+d[:,0]
    CIFs = torch.exp(embeddings)
    return CIFs


def computeSpikeRates(trials_times, spikes_times):
    n_trials = len(spikes_times)
    n_neurons = len(spikes_times[0])
    spikes_rates = torch.empty((n_trials, n_neurons))
    for r in range(n_trials):
        trial_duration = torch.max(trials_times[r])-torch.min(trials_times[r])
        for n in range(n_neurons):
            spikes_rates[r, n] = len(spikes_times[r][n])/trial_duration
    return spikes_rates


def saveDataForMatlabEstimations(qMu, qSVec, qSDiag, C, d,
                                 indPointsLocs,
                                 legQuadPoints, legQuadWeights,
                                 kernelsTypes, kernelsParams,
                                 spikesTimes,
                                 indPointsLocsKMSRegEpsilon, trialsLengths,
                                 latentsTrialsTimes,
                                 emMaxIter, eStepMaxIter, mStepEmbeddingMaxIter,
                                 mStepKernelsMaxIter, mStepIndPointsMaxIter,
                                 saveFilename):
    n_trials = len(spikesTimes)
    nNeurons = len(spikesTimes[0])
    nLatents = len(qMu)
    # indPointsLocsKMSEpsilon = np.array(indPointsLocsKMSEpsilon)
    mdict = dict(n_trials=n_trials, nNeurons=nNeurons, nLatents=nLatents,
                 C=C.numpy(), d=torch.reshape(input=d, shape=(-1,1)).numpy(),
                 legQuadPoints=legQuadPoints.numpy(),
                 legQuadWeights=legQuadWeights.numpy(),
                 indPointsLocsKMSRegEpsilon=indPointsLocsKMSRegEpsilon,
                 trialsLengths=trialsLengths.numpy(),
                 emMaxIter=emMaxIter, eStepMaxIter=eStepMaxIter,
                 mStepEmbeddingMaxIter=mStepEmbeddingMaxIter,
                 mStepKernelsMaxIter=mStepKernelsMaxIter,
                 mStepIndPointsMaxIter=mStepIndPointsMaxIter)
    for k in range(nLatents):
        mdict.update({"kernelType_{:d}".format(k): kernelsTypes[k]})
        mdict.update({"qMu_{:d}".format(k): qMu[k].numpy().astype(np.float64)})
        mdict.update({"qSVec_{:d}".format(k): qSVec[k].numpy().astype(np.float64)})
        mdict.update({"qSDiag_{:d}".format(k): qSDiag[k].numpy().astype(np.float64)})
        mdict.update({"kernelsParams_{:d}".format(k): kernelsParams[k].numpy().astype(np.float64)})
        mdict.update({"indPointsLocs_{:d}".format(k): indPointsLocs[k].numpy().astype(np.float64)})
        mdict.update({"qMu_{:d}".format(k): qMu[k].numpy().astype(np.float64)})
        mdict.update({"qSVec_{:d}".format(k): qSVec[k].numpy().astype(np.float64)})
        mdict.update({"qSDiag_{:d}".format(k): qSDiag[k].numpy().astype(np.float64)})
        mdict.update({"latentsTrialsTimes_{:d}".format(k): latentsTrialsTimes[k].numpy().astype(np.float64)})
    for r in range(n_trials):
        for n in range(nNeurons):
            mdict.update({"spikesTimes_{:d}_{:d}".format(r, n): spikesTimes[r][n].numpy().astype(np.float64)})
    scipy.io.savemat(file_name=saveFilename, mdict=mdict)

def getCholFromVec(vec, nIndPoints):
    chol = torch.zeros((nIndPoints, nIndPoints), dtype=torch.double)
    trilIndices = torch.tril_indices(nIndPoints, nIndPoints)
    chol[trilIndices[0,:],trilIndices[1,:]] = vec
    return chol

def buildCovsFromCholVecs(cholVecs):
    K = len(cholVecs)
    R = cholVecs[0].shape[0]
    covs = [[None] for k in range(K)]
    for k in range(K):
        # cholVecs[k] \in n_trials x Pk x 1
        # Pk = ((nIndPointsK+1)*nIndPointsK)/2
        # nIndPointsK = (-1+sqrt(1+8*Pk))/2
        Pk = cholVecs[k].shape[1]
        nIndPointsK = int((-1+math.sqrt(1+8*Pk))/2)
        covs[k] = torch.empty((R, nIndPointsK, nIndPointsK), dtype=torch.double)
        for r in range(R):
            cholKR = getCholFromVec(vec=cholVecs[k][r,:,0], nIndPoints=nIndPointsK)
            covs[k][r,:,:] = torch.matmul(cholKR, torch.transpose(cholKR, 0, 1))
    return covs


def getQSVecsAndQSDiagsFromQSCholVecs(qsCholVecs):
    # qsCholVecs[k] \in nTrial x Pk
    nLatents = len(qsCholVecs)
    n_trials = qsCholVecs[0].shape[0]
    qSVec = [[] for k in range(nLatents)]
    qSDiag = [[] for k in range(nLatents)]
    for k in range(nLatents):
        Pk = qsCholVecs[k].shape[1]
        nIndPointsK = int((-1.0+math.sqrt(1+8*Pk))/2.0)
        qSVec[k] = torch.empty(n_trials, nIndPointsK, 1, dtype=torch.double)
        qSDiag[k] = torch.empty(n_trials, nIndPointsK, 1, dtype=torch.double)
        for r in range(n_trials):
            qSRSigmaKR = getCholFromVec(vec=qsCholVecs[k][r, :, 0], nIndPoints=nIndPointsK)
            qSigmaKR = torch.matmul(qSRSigmaKR, torch.transpose(qSRSigmaKR, 0, 1))
            qSDiagKR = torch.diag(qSigmaKR)
            qSigmaKR = qSigmaKR - torch.diag(qSDiagKR)
            eValKR, eVecKR = torch.eig(qSigmaKR, eigenvectors=True)
            maxEvalIKR = torch.argmax(eValKR, dim=0)[0]
            qSVecKR = eVecKR[:, maxEvalIKR]*torch.sqrt(eValKR[maxEvalIKR, 0])
            qSVec[k][r, :, 0] = qSVecKR
            qSDiag[k][r, :, 0] = qSDiagKR
    return qSVec, qSDiag


def clock(func):
    def clocked(*args,**kargs):
        t0 = time.perf_counter()
        result = func(*args,**kargs)
        elapsed = time.perf_counter()-t0
        name = func.__name__
        if len(args)>0:
            arg_str = ', '.join(repr(arg) for arg in args)
        else:
            arg_str = None
        if len(kargs)>0:
            keys = kargs.keys()
            values = kargs.values()
            karg_str = ', '.join(key + "=" + repr(value) for key in keys for value in values)
        else:
            karg_str = None
        if arg_str is not None and karg_str is not None:
            print('[%0.8fs] %s(%s,%s) -> %r' % (elapsed, name, arg_str, karg_str, result))
        elif arg_str is not None:
            print('[%0.8fs] %s(%s) -> %r' % (elapsed, name, arg_str, result))
        elif karg_str is not None:
            print('[%0.8fs] %s(%s) -> %r' % (elapsed, name, karg_str, result))
        else:
            print('[%0.8fs] %s() -> %r' % (elapsed, name, result))
        return result
    return clocked

def chol3D(K):
#     if my_globals.raise_exception:
#         raise ValueError("Test error in chol3D")
    Kchol = torch.zeros(K.shape, dtype=K.dtype, device=K.device)
    nTrial = K.shape[0]
    for i in range(nTrial):
        Kchol[i,:,:] = torch.linalg.cholesky(K[i,:,:])
    return Kchol

def pinv3D(K, rcond=1e-15):
    Kpinv = torch.zeros(K.shape, dtype=K.dtype, device=K.device)
    nTrial = K.shape[0]
    for i in range(nTrial):
        Kpinv[i,:,:] = torch.linalg.pinv(K[i,:,:], rcond=rcond)
    return Kpinv


def getLegQuadPointsAndWeights(n_quad, trials_start_times, trials_end_times,
                               dtype=torch.double):
    n_trials = len(trials_start_times)
    assert(n_trials == len(trials_end_times))
    leg_quad_points = torch.empty((n_trials, n_quad, 1), dtype=dtype)
    leg_quad_weights = torch.empty((n_trials, n_quad, 1), dtype=dtype)
    for r in range(n_trials):
        leg_quad_points[r, :, 0], leg_quad_weights[r, :, 0] = \
                gcnu_common.numerical_methods.utils.leggaussVarLimits(
                    n=n_quad, a=trials_start_times[r], b=trials_end_times[r])
    return leg_quad_points, leg_quad_weights


def getTrialsTimes(start_times, end_times, n_steps):
    assert(len(start_times) == len(end_times))
    n_trials = len(start_times)
    trials_times = torch.empty((n_trials, n_steps, 1), dtype=torch.double)
    for r in range(n_trials):
        trials_times[r, :, 0] = torch.linspace(start_times[r], end_times[r],
                                               n_steps)
    return trials_times


def computeSpikeClassificationROC(spikes_times, cif_times, cif_values,
                                  highres_bin_size=1e-3):
    f = scipy.interpolate.interp1d(cif_times, cif_values)
    cif_times_highres = np.arange(cif_times[0], cif_times[-1],
                                  highres_bin_size)
    cif_values_highres = f(cif_times_highres)
    bins = pd.interval_range(start=cif_times[0].item(),
                             end=cif_times[-1].item(),
                             periods=len(cif_times_highres))
    cut_res = pd.cut(spikes_times, bins=bins, retbins=False)
    Y = cut_res.value_counts().values
    indicesMoreThanOneSpikes = (Y > 1).nonzero()
    if len(indicesMoreThanOneSpikes) > 0:
        warnings.warn("Found more than one spike in {:d} bins".format(
            len(indicesMoreThanOneSpikes)))
        Y[indicesMoreThanOneSpikes] = 1.0
    fpr, tpr, thresholds = sklearn.metrics.roc_curve(Y, cif_values_highres,
                                                     pos_label=1)
    roc_auc = sklearn.metrics.auc(fpr, tpr)
    return fpr, tpr, roc_auc


def getLatentsMeansAndSTDs(meansFuncs, kernels, trialsTimes):
    n_trials = len(trialsTimes)
    nLatents = len(kernels)
    latents_means = [[] for r in range(n_trials)]
    latents_STDs = [[] for r in range(n_trials)]

    for r in range(n_trials):
        latents_means[r] = torch.empty((nLatents, len(trialsTimes[r])))
        latents_STDs[r] = torch.empty((nLatents, len(trialsTimes[r])))
        for k in range(nLatents):
            gp = gcnu_common.stats.gaussianProcesses.eval.GaussianProcess(mean=meansFuncs[k], kernel=kernels[k])
            latents_means[r][k,:] = gp.mean(t=trialsTimes[r])
            latents_STDs[r][k,:] = gp.std(t=trialsTimes[r])
    return latents_means, latents_STDs

def getLatentsSTDs(kernels, trialsTimes):
    n_trials = len(trialsTimes)
    nLatents = len(kernels)
    latents_STDs = [[] for r in range(n_trials)]

    for r in range(n_trials):
        latents_STDs[r] = torch.empty((nLatents, len(trialsTimes[r])))
        for k in range(nLatents):
            latents_STDs[r][k,:] = kernels[k].buildKernelMatrixDiag(X=trialsTimes[r]).sqrt()
    return latents_STDs

# def getLatentsMeanFuncsSamples(latents_meansFuncs, trialsTimes, dtype):
#     n_trials = len(latents_meansFuncs)
#     nLatents = len(latents_meansFuncs[0])
#     latents_meansFuncsSamples = [[] for r in range(n_trials)]
#     for r in range(n_trials):
#         latents_meansFuncsSamples[r] = torch.empty((nLatents, len(trialsTimes[r])), dtype=dtype)
#         for k in range(nLatents):
#             latents_meansFuncsSamples[r][k,:] = latents_meansFuncs[r][k](t=trialsTimes[r])

def getLatentsSamplesMeansAndSTDsFromSampledMeans(n_trials, sampledMeans, kernels, trialsTimes, latentsGPRegularizationEpsilon, dtype):
    nLatents = len(kernels)
    latents_samples = [[] for r in range(n_trials)]
    latents_means = [[] for r in range(n_trials)]
    latents_STDs = [[] for r in range(n_trials)]

    for r in range(n_trials):
        latents_samples[r] = torch.empty((nLatents, len(trialsTimes[r])), dtype=dtype)
        latents_means[r] = torch.empty((nLatents, len(trialsTimes[r])), dtype=dtype)
        latents_STDs[r] = torch.empty((nLatents, len(trialsTimes[r])), dtype=dtype)
        for k in range(nLatents):
            print("Procesing trial {:d} and latent {:d}".format(r+1, k+1))
            mean = sampledMeans[r,:,k]
            cov = kernels[k].buildKernelMatrix(trialsTimes[r])
            cov = cov + latentsGPRegularizationEpsilon*torch.eye(cov.shape[0])
            std = torch.diag(cov).sqrt()
            mn = scipy.stats.multivariate_normal(mean=mean, cov=cov)
            sample = torch.from_numpy(mn.rvs())
            latents_samples[r][k,:] = sample
            latents_means[r][k,:] = mean
            latents_STDs[r][k,:] = std
            # plt.plot(trialsTimes[r], mean, label="mean")
            # plt.plot(trialsTimes[r], sample, label="sample")
            # plt.xlabel("Time (sec)")
            # plt.ylabel("Value")
            # plt.title("Latent {:d}".format(k))
            # plt.legend()
            # plt.show()
    return latents_samples, latents_means, latents_STDs

def getDiagIndicesIn3DArray(N, M, device=torch.device("cpu")):
    frameDiagIndices = torch.arange(end=N, device=device)*(N+1)
    frameStartIndices = torch.arange(end=M, device=device)*N**2
    # torch way of computing an outer sum
    diagIndices = (frameDiagIndices.reshape(-1,1)+frameStartIndices).flatten()
    answer, _ = diagIndices.sort()
    return answer

def build3DdiagFromDiagVector(v, N, M):
    assert(len(v)==N*M)
    diagIndices = getDiagIndicesIn3DArray(N=N, M=M)
    D = torch.zeros(M*N*N, dtype=v.dtype, device=v.device)
    D[diagIndices] = v
    reshapedD = D.reshape(shape = (M, N, N))
    return reshapedD

def buildQSigmaFromQSVecAndQSDiag(qSVec, qSDiag):
    nLatents = len(qSVec)
    n_trials = qSVec[0].shape[0]
    qSigma = [[None] for k in range(nLatents)]
    for k in range(nLatents):
        nIndK = qSDiag[k].shape[1]
        # qq \in n_trials x nInd[k] x 1
        qq = qSVec[k].reshape(shape=(n_trials, nIndK, 1))
        # dd \in n_trials x nInd[k] x 1
        nIndKVarRnkK = qSVec[k].shape[1]
        dd = build3DdiagFromDiagVector(v=(qSDiag[k].flatten())**2, M=n_trials, N=nIndKVarRnkK)
        # qSigma[k] \in n_trials x nInd[k] x nInd[k]
        qSigma[k] = torch.matmul(qq, torch.transpose(a=qq, dim0=1, dim1=2)) + dd
    return qSigma

def getSRQSigmaVec(qSVec, qSDiag):
    nLatents = len(qSVec)
    n_trials = qSVec[0].shape[0]
    qSigma = buildQSigmaFromQSVecAndQSDiag(qSVec=qSVec, qSDiag=qSDiag)
    srQSigmaVec = [[None] for k in range(nLatents)]
    for k in range(nLatents):
        nIndPointsK = qSigma[k].shape[1]
        Pk = int((nIndPointsK+1)*nIndPointsK/2)
        srQSigmaVec[k] = torch.empty((n_trials, Pk, 1), dtype=torch.double)
        for r in range(n_trials):
            cholKR = torch.linalg.cholesky(qSigma[k][r,:,:])
            trilIndices = torch.tril_indices(nIndPointsK, nIndPointsK)
            cholKRVec = cholKR[trilIndices[0,:], trilIndices[1,:]]
            srQSigmaVec[k][r,:,0] = cholKRVec
    return srQSigmaVec

# def getIndPointLocs0(nIndPointsPerLatent, trialsLengths, firstIndPointLoc):
#     nLatents = len(nIndPointsPerLatent)
#     n_trials = len(trialsLengths)
#
#     Z0 = [[] for k in range(nLatents)]
#     for k in range(nLatents):
#         Z0[k] = torch.empty((n_trials, nIndPointsPerLatent[k], 1), dtype=torch.double)
#         for r in range(n_trials):
#             Z0[k][r,:,0] = torch.linspace(firstIndPointLoc, trialsLengths[r], nIndPointsPerLatent[k])
#     return Z0


def getEmbeddingSamples(C, d, latents_samples):
    n_trials = len(latents_samples)
    answer = [torch.matmul(C, latents_samples[r])+d for r in range(n_trials)]
    return answer


def getEmbeddingMeans(C, d, latents_means):
    n_trials = len(latents_means)
    answer = [torch.matmul(C, latents_means[r])+d for r in range(n_trials)]
    return answer


def getEmbeddingSTDs(C, latents_STDs):
    n_trials = len(latents_STDs)
    answer = [torch.matmul(C**2, latents_STDs[r]**2).sqrt()
              for r in range(n_trials)]
    return answer


def getSRQSigmaVecsFromKzz(Kzz):
    Kzz_chol = []
    for aKzz in Kzz:
        Kzz_chol.append(svGPFA.utils.miscUtils.chol3D(aKzz))
    answer = getVectorRepOfLowerTrianMatrices(lt_matrices=Kzz_chol)
    return answer


def getVectorRepOfLowerTrianMatrices(lt_matrices):
    """Returns vectors containing the lower-triangular elements of the input lower-triangular matrices.

    :parameter lt_matrices: a list of length n_latents, with lt_matrices[k] a tensor of dimension n_trials x n_ind_points x n_ind_points, where lt_matrices[k][r, :, :] is a lower-triangular matrix.
    :type lt_matrices: list

    :return: a list vec_lt_matrices of length n_latents, whith vec_lt_matrices[k] a tensor of dimension (n_trials, n_ind_points*(n_ind_points+1)/2, 0), where vec_lt_matrices[k][r, :, 0] contains the vectorized lower-triangular elements of lt_matrices[k][r, :, :].

    """

    n_latents = len(lt_matrices)
    n_trials = lt_matrices[0].shape[0]

    srQSigmaVec = [[None] for k in range(n_latents)]
    for k in range(n_latents):
        nIndPointsK = lt_matrices[k].shape[1]
        Pk = int((nIndPointsK+1)*nIndPointsK/2)
        srQSigmaVec[k] = torch.empty((n_trials, Pk, 1), dtype=torch.double)
        for r in range(n_trials):
            cholKR = lt_matrices[k][r, :, :]
            trilIndices = torch.tril_indices(nIndPointsK, nIndPointsK)
            cholKRVec = cholKR[trilIndices[0, :], trilIndices[1, :]]
            srQSigmaVec[k][r, :, 0] = cholKRVec
    return srQSigmaVec







import pdb
import torch
import abc
import svGPFA.utils.miscUtils


class KernelsMatricesStore(abc.ABC):

    @abc.abstractmethod
    def buildKernelsMatrices(self):
        pass

    def setKernels(self, kernels):
        self._kernels = kernels

    def setInitialParams(self, initial_params):
        self.setIndPointsLocs(
            ind_points_locs=initial_params["inducing_points_locs0"])
        self.setKernelsParams(kernels_params=initial_params["kernels_params0"])

    def setKernelsParams(self, kernels_params):
        for k in range(len(self._kernels)):
            self._kernels[k].setParams(kernels_params[k])

    def setIndPointsLocs(self, ind_points_locs):
        self._ind_points_locs = ind_points_locs

    def getIndPointsLocs(self):
        return self._ind_points_locs

    def getKernels(self):
        return self._kernels

    def getKernelsParams(self):
        answer = []
        for i in range(len(self._kernels)):
            answer.append(self._kernels[i].getParams())
        return answer


class KernelMatricesStoreGettersAndSetters(abc.ABC):
    def get_flattened_kernels_params(self):
        flattened_params = []
        for k in range(len(self._kernels)):
            flattened_params.extend(self._kernels[k].getParams().flatten().tolist())
        return flattened_params

    def get_flattened_kernels_params_grad(self):
        flattened_params_grad = []
        for k in range(len(self._kernels)):
            flattened_params_grad.extend(self._kernels[k].getParams().grad.flatten().tolist())
        return flattened_params_grad

    def set_kernels_params_from_flattened(self, flattened_params):
        for k in range(len(self._kernels)):
            kernel_nParams = self._kernels[k].getParams().numel()
            flattened_param = flattened_params[:kernel_nParams]
            self._kernels[k].setParams(torch.tensor(flattened_param, dtype=torch.double))
            flattened_params = flattened_params[kernel_nParams:]

    def set_kernels_params_requires_grad(self, requires_grad):
        for k in range(len(self._kernels)):
            self._kernels[k].getParams().requires_grad = requires_grad

    def get_flattened_ind_points_locs(self):
        flattened_params = []
        for k in range(len(self._ind_points_locs)):
            flattened_params.extend(self._ind_points_locs[k].flatten().tolist())
        return flattened_params

    def get_flattened_ind_points_locs_grad(self):
        flattened_params_grad = []
        for k in range(len(self._ind_points_locs)):
            flattened_params_grad.extend(self._ind_points_locs[k].grad.flatten().tolist())
        return flattened_params_grad

    def set_ind_points_locs_from_flattened(self, flattened_params):
        for k in range(len(self._ind_points_locs)):
            numel = self._ind_points_locs[k].numel()
            self._ind_points_locs[k] = torch.tensor(flattened_params[:numel],
                                                    dtype=torch.double).reshape(self._ind_points_locs[k].shape)
            flattened_params = flattened_params[numel:]

    def set_ind_points_locs_requires_grad(self, requires_grad):
        for k in range(len(self._ind_points_locs)):
            self._ind_points_locs[k].requires_grad = requires_grad


class IndPointsLocsKMS(KernelsMatricesStore):

    def setRegParam(self, reg_param):
        self._reg_param = reg_param

    # @abc.abstractmethod
    def _invertKzz3D(self, Kzz):
        pass

    @abc.abstractmethod
    def solveForLatent(self, input, latentIndex):
        pass

    @abc.abstractmethod
    def solveForLatentAndTrial(self, input, latentIndex, trialIndex):
        pass

    def buildKernelsMatrices(self):
        n_latents = len(self._kernels)
        self._Kzz = [[None] for k in range(n_latents)]
        self._Kzz_inv = [[None] for k in range(n_latents)]

        for k in range(n_latents):
            self._Kzz[k] = (self._kernels[k].buildKernelMatrix(X1=self._ind_points_locs[k])+
                            self._reg_param*torch.eye(n=self._ind_points_locs[k].shape[1],
                                                      dtype=self._ind_points_locs[k].dtype,
                                                      device=self._ind_points_locs[k].device))
            self._Kzz_inv[k] = self._invertKzz3D(self._Kzz[k]) # O(n^3)

    def getKzz(self):
        return self._Kzz

    def getRegParam(self):
        return self._reg_param


class IndPointsLocsKMS_Chol(IndPointsLocsKMS):

    def _invertKzz3D(self, Kzz):
        Kzz_inv = svGPFA.utils.miscUtils.chol3D(Kzz)  # O(n^3)
        return Kzz_inv

    def solveForLatent(self, input, latentIndex):
        solve = torch.cholesky_solve(input, self._Kzz_inv[latentIndex])
        return solve

    def solveForLatentAndTrial(self, input, latentIndex, trialIndex):
        solve = torch.cholesky_solve(input, self._Kzz_inv[latentIndex][trialIndex, :, :])
        return solve


class IndPointsLocsKMS_CholWithGettersAndSetters(IndPointsLocsKMS_Chol, KernelMatricesStoreGettersAndSetters):
    def __init__(self):
        pass


class IndPointsLocsKMS_PInv(IndPointsLocsKMS):

    def _invertKzz3D(self, Kzz):
        Kzz_inv = svGPFA.utils.miscUtils.pinv3D(Kzz)  # O(n^3)
        return Kzz_inv

    def solveForLatent(self, input, latentIndex):
        solve = torch.matmul(self._Kzz_inv[latentIndex], input)
        return solve

    def solveForLatentAndTrial(self, input, latentIndex, trialIndex):
        solve = torch.matmul(self._Kzz_inv[latentIndex][trialIndex, :, :],
                             input)
        return solve


class IndPointsLocsKMS_PInvWithGettersAndSetters(IndPointsLocsKMS_PInv,
                                                 KernelMatricesStoreGettersAndSetters):
    def __init__(self):
        pass


class IndPointsLocsAndTimesKMS(KernelsMatricesStore):

    def setTimes(self, times):
        # times \in nTrials x nQuad x 1
        self._t = times

    def getKtz(self):
        return self._Ktz

    def getKtt(self):
        return self._Ktt

    def getKttDiag(self):
        return self._KttDiag


class IndPointsLocsAndAllTimesKMS(IndPointsLocsAndTimesKMS):

    def buildKernelsMatrices(self):
        # self._t \in nTrials x nQuad x 1
        n_latents = len(self._ind_points_locs)
        self._Ktz = [[None] for k in range(n_latents)]
        self._KttDiag = torch.zeros(self._t.shape[0], self._t.shape[1],
                                    n_latents,
                                    dtype=self._t.dtype, device=self._t.device)
        for k in range(n_latents):
            self._Ktz[k] = self._kernels[k].buildKernelMatrix(X1=self._t, X2=self._ind_points_locs[k])
            self._KttDiag[:, :, k] = self._kernels[k].buildKernelMatrixDiag(X=self._t).squeeze()

    def buildKttKernelsMatrices(self):
        # t \in nTrials x nQuad x 1
        n_latents = len(self._ind_points_locs)
        self._Ktt = [[None] for k in range(n_latents)]

        for k in range(n_latents):
            self._Ktt[k] = self._kernels[k].buildKernelMatrix(X1=self._t, X2=self._t)


class IndPointsLocsAndAssocTimesKMS(IndPointsLocsAndTimesKMS):

    def buildKernelsMatrices(self):
        n_latents = len(self._ind_points_locs)
        n_trials = self._ind_points_locs[0].shape[0]
        self._Ktz = [[[None] for tr in range(n_trials)]
                     for k in range(n_latents)]
        self._KttDiag = [[[None] for tr in range(n_trials)] for k in
                         range(n_latents)]

        for k in range(n_latents):
            for tr in range(n_trials):
                self._Ktz[k][tr] = self._kernels[k].buildKernelMatrix(
                    X1=self._t[tr], X2=self._ind_points_locs[k][tr, :, :])
                self._KttDiag[k][tr] = self._kernels[k].buildKernelMatrixDiag(
                    X=self._t[tr])



import torch

def getTrueVariationalMean(t, latent_mean, inducing_points_locs, kernel):
    Ktz = kernel.buildKernelMatrix(X1=t, X2=inducing_points_locs)
    Kzz = kernel.buildKernelMatrix(X1=inducing_points_locs, X2=inducing_points_locs)
    res = torch.linalg.lstsq(Ktz, latent_mean)
    vMean = torch.matmul(Kzz, res.solution)
    return vMean
